use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007006
use Test::Spelling 0.17;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
Anwar
AuthPassword
AuthUsername
BacklogInterval
BatchDir
BatchDiskSpace
Brad
CleanupInterval
Component
DefaultIndex
DefaultType
ElasticSearch
Elasticsearch
FlushInterval
FlushSize
Indexer
KeepAliveTimeout
Lhotsky
LoggingConfig
MaxConnsPerServer
MaxConnsTotal
MaxFailedRatio
MaxPendingRequests
MaxRecoveryBatches
Mohammad
POE
PoolConnections
StatsHandler
StatsInterval
bin
brad
elasticsearch
es
file
ie
lib
mohammad
